/*	Wavefront OBJ file loader
	
	adapted from the GLM library (distributed with GLUT) Written by Nate Robins, 1997.
 	email: ndr@pobox.com www: http://www.pobox.com/~ndr
 	
 	Modifications, 'Mac-ifying' and 'Quake-ifying' by Frank Condello
 	email: pox@planetquake.com	www: http://www.planetquake.com/paroxysm/
 	
 	Changes from GLM:
	- Support for n-sided polygons (up to 32 verts)
	- Request polygon type at loadtime (tris, quads, all)
	- No 'default' group for properly formated OBJ files
	- No material library parsing (material name saved as texture reference)
	- Save Creator info (assumes first comment contains creator info)
*/

#define OBJ_MAX_POLYVERTS 	32

// 3 choices when loading models:
// OBJ_TRIANGLES only loads fully triangulated models (error if not)
// OBJ_TRIS_AND_QUADS loads models with tris and quads (error if not)
// OBJ_NSIDED loads models with tris, quads, and n-sided polys (error if n>OBJ_MAX_POLYVERTS)
#define U_OBJ_NSIDED		4	// Don't use directly!
#define U_OBJ_QUADS			2	// Don't use directly!
#define OBJ_TRIANGLES		1
#define OBJ_TRIS_AND_QUADS	U_OBJ_QUADS+OBJ_TRIANGLES
#define OBJ_NSIDED			OBJ_TRIS_AND_QUADS+U_OBJ_NSIDED

#define T(x) model->polys[(x)]
#define MOVEPTR(x) ptr++; (x>9)?ptr++:0; (x>99)?ptr++:0; (x>999)?ptr++:0; (x>9999)?ptr++:0;

enum { X, Y, Z, W };

// general purpose node structure
typedef struct node_s
{
	unsigned 		index;
	qboolean		averaged;
	struct node_s	*next;
}
node_t;


//
// OBJpoly_t: Structure that defines a polygon in a model.
//
typedef struct
{
	unsigned	vindices[OBJ_MAX_POLYVERTS];	// array of vertex indices
	unsigned	nindices[OBJ_MAX_POLYVERTS];	// array of normal indices
	unsigned	tindices[OBJ_MAX_POLYVERTS];	// array of texcoord indices
	unsigned	findex;							// index of facet normal
	unsigned	numverts;						// number of vertecies in this poly
}
OBJpoly_t;


//
// OBJgroup: Structure that defines a group in a model.
//
typedef struct OBJgroup_s
{
	char				*name;		// name of this group
	char				*texname;   // texture name for this group
	unsigned			numpolys;	// number of polys in this group
	unsigned			*polys;		// array of polygon indices
	struct OBJgroup_s	*next;		// pointer to next group in model
}
OBJgroup_t;


//
// OBJmodel: Structure that defines a model.
//
typedef struct
{
	char			*name;			// path to this model
	char			*creator;		// app that produced this model
	
	unsigned	  	numvertices;	// number of vertices in model
	float			*vertices;		// array of vertices 

	unsigned		numnormals;		// number of normals in model
	float			*normals;		// array of normals

	unsigned		numtexcoords;	// number of texcoords in model
	float			*texcoords;		// array of texture coordinates

	unsigned		numfacetnorms;	// number of facetnorms in model
	float			*facetnorms;	// array of facetnorms

	unsigned		numpolys;		// number of polygons in model
	OBJpoly_t		*polys;			// array of polygons

	unsigned		numgroups;		// number of groups in model
	OBJgroup_t		*groups;		// linked list of groups
	
	unsigned		maxnormals;		// Number of potential normals (before numnormals is calculated)
}
OBJmodel_t;


OBJmodel_t*	LoadOBJModel (char *filename, unsigned flags, float smooth, vec3_t translate);
void 		FreeOBJModel (OBJmodel_t *model);