
#include "bsp5.h"

// pOx - Multi-WAD
#include "Win32.h"
extern 	WIN32_FIND_DATA ff;
extern 	HANDLE ffhandle;
extern	char		wadfilename[1024];

int		headclipnode;
int		firstface;

//===========================================================================

/*
==================
FindFinalPlane

Used to find plane index numbers for clip nodes read from child processes
==================
*/
int FindFinalPlane (dplane_t *p)
{
	int		i;
	dplane_t	*dplane;
	
	for (i=0, dplane = dplanes ; i<numplanes ; i++, dplane++)
	{
		if (p->type != dplane->type)
			continue;
		if (p->dist != dplane->dist)
			continue;
		if (p->normal[0] != dplane->normal[0])
			continue;
		if (p->normal[1] != dplane->normal[1])
			continue;
		if (p->normal[2] != dplane->normal[2])
			continue;
		return i;
	}
	
//
// new plane
//
	if (numplanes == MAX_MAP_PLANES)
		Error ("numplanes == MAX_MAP_PLANES");
	dplane = &dplanes[numplanes];
	*dplane = *p;
	numplanes++;
	
	return numplanes - 1;
}



int		planemapping[MAX_MAP_PLANES];

void WriteNodePlanes_r (node_t *node)
{
	plane_t		*plane;
	dplane_t	*dplane;

	if (node->planenum == -1)
		return;
	if (planemapping[node->planenum] == -1)
	{	// a new plane
		planemapping[node->planenum] = numplanes;
		
		if (numplanes == MAX_MAP_PLANES)
			Error ("numplanes == MAX_MAP_PLANES");
		plane = &planes[node->planenum];
		dplane = &dplanes[numplanes];
		dplane->normal[0] = plane->normal[0];
		dplane->normal[1] = plane->normal[1];
		dplane->normal[2] = plane->normal[2];
		dplane->dist = plane->dist;
		dplane->type = plane->type;

		numplanes++;
	}

	node->outputplanenum = planemapping[node->planenum];
	
	WriteNodePlanes_r (node->children[0]);
	WriteNodePlanes_r (node->children[1]);
}

/*
==================
WriteNodePlanes

==================
*/
void WriteNodePlanes (node_t *nodes)
{
	memset (planemapping,-1, sizeof(planemapping));
	WriteNodePlanes_r (nodes);
}

//===========================================================================

/*
==================
WriteClipNodes_r

==================
*/
int WriteClipNodes_r (node_t *node)
{
	int			i, c;
	dclipnode_t	*cn;
	int			num;
	
// FIXME: free more stuff?	
	if (node->planenum == -1)
	{
		num = node->contents;
		free (node);
		return num;
	}
	
// emit a clipnode
	c = numclipnodes;
	cn = &dclipnodes[numclipnodes];
	numclipnodes++;
	cn->planenum = node->outputplanenum;
	for (i=0 ; i<2 ; i++)
		cn->children[i] = WriteClipNodes_r(node->children[i]);
	
	free (node);
	return c;
}

/*
==================
WriteClipNodes

Called after the clipping hull is completed.  Generates a disk format
representation and frees the original memory.
==================
*/
void WriteClipNodes (node_t *nodes)
{
	headclipnode = numclipnodes;
	WriteClipNodes_r (nodes);
}

//===========================================================================


#ifdef DEBUG
static int face_ref_freq[MAX_MAP_MARKSURFACES];
#endif

/*
==================
WriteLeaf
==================
*/
void WriteLeaf (node_t *node)
{
	face_t		**fp, *f;
	dleaf_t		*leaf_p;
		
// emit a leaf
	leaf_p = &dleafs[numleafs];
	numleafs++;

	leaf_p->contents = node->contents;

//
// write bounding box info
//	
	VectorCopy (node->mins, leaf_p->mins);
	VectorCopy (node->maxs, leaf_p->maxs);
	
	leaf_p->visofs = -1;	// no vis info yet
	
//
// write the marksurfaces
//
	leaf_p->firstmarksurface = nummarksurfaces;
	
	for (fp=node->markfaces ; *fp ; fp++)
	{
	// emit a marksurface
		if (nummarksurfaces == MAX_MAP_MARKSURFACES)
			Error ("nummarksurfaces == MAX_MAP_MARKSURFACES");
		f = *fp;
      if (f->texturenum<0) continue;
		do
		{
         if (f->outputnumber<0) Error("f->outputnumber<0");
#ifdef DEBUG
         face_ref_freq[f->outputnumber]++;
#endif
			dmarksurfaces[nummarksurfaces] =  f->outputnumber;
			nummarksurfaces++;
			f=f->original;		// grab tjunction split faces
		} while (f);
	}
	
	leaf_p->nummarksurfaces = nummarksurfaces - leaf_p->firstmarksurface;
}


/*
==================
WriteDrawNodes_r
==================
*/
void WriteDrawNodes_r (node_t *node)
{
	dnode_t	*n;
	int		i;

// emit a node	
	if (numnodes == MAX_MAP_NODES)
		Error ("numnodes == MAX_MAP_NODES");
	n = &dnodes[numnodes];
	numnodes++;

	VectorCopy (node->mins, n->mins);
	VectorCopy (node->maxs, n->maxs);

	n->planenum = node->outputplanenum;
	n->firstface = node->firstface;
	n->numfaces = node->numfaces;

//
// recursively output the other nodes
//	
	
	for (i=0 ; i<2 ; i++)
	{
		if (node->children[i]->planenum == -1)
		{
			if (node->children[i]->contents == CONTENTS_SOLID)
				n->children[i] = -1;
			else
			{
				n->children[i] = -(numleafs + 1);
				WriteLeaf (node->children[i]);
			}
		}
		else
		{
			n->children[i] = numnodes;	
			WriteDrawNodes_r (node->children[i]);
		}
	}
}

/*
==================
WriteDrawNodes
==================
*/
void WriteDrawNodes (node_t *headnode)
{
	int		i;
	int		start;
	dmodel_t	*bm;

#if 0
	if (headnode->contents < 0)
		Error ("FinishBSPModel: empty model");
#endif

// emit a model
	if (nummodels == MAX_MAP_MODELS)
		Error ("nummodels == MAX_MAP_MODELS");
	bm = &dmodels[nummodels];
	nummodels++;
	
	bm->headnode[0] = numnodes;
	bm->firstface = firstface;
	bm->numfaces = numfaces - firstface;	
	firstface = numfaces;
	
	start = numleafs;

	if (headnode->contents < 0)	
		WriteLeaf (headnode);
	else
		WriteDrawNodes_r (headnode);
	bm->visleafs = numleafs - start;
	
	for (i=0 ; i<3 ; i++)
	{
		bm->mins[i] = headnode->mins[i] + SIDESPACE + 1;	// remove the padding
		bm->maxs[i] = headnode->maxs[i] - SIDESPACE - 1;
	}
// FIXME: are all the children decendant of padded nodes?

#ifdef DEBUG
   {
      int i;

      for (i=0;i<numfaces;i++)
      {
         if (face_ref_freq[i]!=1)
            printf("face_ref_freq[%i]=%i\n",i,face_ref_freq[i]);
      }
   }
#endif
}


/*
==================
BumpModel

Used by the clipping hull processes that only need to store headclipnode
==================
*/
void BumpModel (int hullnum)
{
	dmodel_t	*bm;

// emit a model
	if (nummodels == MAX_MAP_MODELS)
		Error ("nummodels == MAX_MAP_MODELS");
	bm = &dmodels[nummodels];
	nummodels++;
	
	bm->headnode[hullnum] = headclipnode;
}

//=============================================================================

typedef struct
{
	char		identification[4];		// should be WAD2
	int			numlumps;
	int			infotableofs;
} wadinfo_t;


typedef struct
{
	int			filepos;
	int			disksize;
	int			size;					// uncompressed
	char		type;
	char		compression;
	char		pad1, pad2;
	char		name[16];				// must be null terminated
} lumpinfo_t;

FILE		*texfile;
wadinfo_t	wadinfo;
lumpinfo_t	*lumpinfo;

void CleanupName (char *in, char *out)
{
	int		i = 0;
	char 	*s = in;
	
	while (*s)
	{
		out[i++] = tolower (*s);
		s ++;
	}
	
	out[i] = '\0';
/*
	int		i;
	
	for (i=0 ; i< 16 ; i++ )
	{
		if (!in[i])
			break;
		out[i] = tolower(in[i]);
	}

	for ( ; i< 16 ; i++ )
		out[i] = 0;
*/
}

/*
=================
TEX_InitFromWad
=================
*/
void	TEX_InitFromWad (char *path)
{
	int			i;
	
	texfile = SafeOpenRead (path);
	SafeRead (texfile, &wadinfo, sizeof(wadinfo));
	if (strncmp (wadinfo.identification, "WAD2", 4))
		Error ("TEX_InitFromWad: %s isn't a wadfile",path);
	wadinfo.numlumps = LittleLong(wadinfo.numlumps);
	wadinfo.infotableofs = LittleLong(wadinfo.infotableofs);
	fseek (texfile, wadinfo.infotableofs, SEEK_SET);
	lumpinfo = malloc(wadinfo.numlumps*sizeof(lumpinfo_t));

	SafeRead (texfile, lumpinfo, wadinfo.numlumps*sizeof(lumpinfo_t));

	for (i=0 ; i<wadinfo.numlumps ; i++)
	{
		CleanupName (lumpinfo[i].name, lumpinfo[i].name);
		lumpinfo[i].filepos = LittleLong(lumpinfo[i].filepos);
		lumpinfo[i].disksize = LittleLong(lumpinfo[i].disksize);
	}
}

/*
==================
LoadLump
==================
*/
int LoadLump (char *name, byte *dest)
{
	int		i;
	char	cname[64];
	
	CleanupName (name, cname);
	
	for (i=0 ; i<wadinfo.numlumps ; i++)
	{
		if (!strcmp(cname, lumpinfo[i].name))
		{
			fseek (texfile, lumpinfo[i].filepos, SEEK_SET);
			
			// pOx - only include texture headers if -nogfx is used
			if (nogfx)
			{
				int			j;
				miptex_t	*miptex = (miptex_t *)dest;
				
				SafeRead (texfile, dest, sizeof(miptex_t) );
				CleanupName ((char *)dest, (char *)dest);
				
				for( j=0; j<MIPLEVELS; j++ )
					miptex->offsets[j] = 0;
				
				return sizeof(miptex_t);
			
			} else	// Textures included in the BSP
			{
				SafeRead (texfile, dest, lumpinfo[i].disksize);
				CleanupName ((char *)dest, (char *)dest);	// 1999-04-28  Correction to make names lower case as in the original bsps by Maddes	
				return lumpinfo[i].disksize;
			}
		}
	}
	
//	printf ("WARNING: texture %s not found\n", name);
	return 0;
}


/*
==================
AddAnimatingTextures
==================
*/
// pOx - Multiple wad support
void AddAnimatingTextures (void)
{
	int			base;
	int			i, j, k;
	char		name[64];
	char		fullpath[1024];
	qboolean	ran_out;

	base = nummiptex;
	
	for (i=0 ; i<base ; i++)
	{
		if (miptex[i][0] != '+')
			continue;
		
		strcpy (name, miptex[i]);
		
		ran_out = false;
		
		// Get the first wad
		sprintf(wadfilename, "textures/*.wad");
		ffhandle = FindFirstFile(wadfilename, &ff);
		if (!ff.cFileName[0])
			Error ("qbsp: No wad files could be found.");
		sprintf(wadfilename, "textures/%s", ff.cFileName);
		
		do 
		{
			strcpy(fullpath, wadfilename);
			
			TEX_InitFromWad (fullpath);
			
			if (verbose)
				printf ("Looking for texture frames in wad %s...\n", fullpath);
			
			for (j=0 ; j<20 ; j++)
			{
				if (j < 10)
					name[1] = '0'+j;
				else
					name[1] = 'A'+j-10;		// alternate animation
			
				CleanupName (name, name);
				
				// see if this name exists in the wadfile
				for (k=0 ; k<wadinfo.numlumps ; k++) {
				
					if (!strcmp(name, lumpinfo[k].name))
					{
						FindMiptex (name);	// add to the miptex list
						if (verbose)
							printf ("Added texture frame %s, from wad %s\n", name, fullpath);
						break;
					}
				}
			}
			
			if (!FindNextFile(ffhandle, &ff))
				ran_out = true;
			else
				sprintf(wadfilename, "textures/%s", ff.cFileName);
		
		} while (!ran_out);
			
	}
	
	if ((nummiptex - base) > 2)
		printf ("Added %i texture frames\n\n", nummiptex - (base-1));
	else
		printf ("No texture frames added\n\n");
}

/*
==================
WriteMiptex
==================
*/
// pOx - Multiple wad support (pulled a boner in r1 - this is a re-write)
void WriteMiptex (void)
{
	int		i, len;
	byte	*data;
	dmiptexlump_t	*l;
	char	fullpath[1024];
	qboolean	missing_texture;
	qboolean	ran_out;
		
	AddAnimatingTextures ();
	
	l = (dmiptexlump_t *)dtexdata;
	data = (byte *)&l->dataofs[nummiptex];
	l->nummiptex = nummiptex;
	
	// Search the wads until the texture is found or we run out of wads
	for (i=0 ; i<nummiptex ; i++)
	{
		ran_out = false;
		
		// Get the first wad
		sprintf(wadfilename, "textures/*.wad");
		ffhandle = FindFirstFile(wadfilename, &ff);
		if (!ff.cFileName[0])
			Error ("WriteMiptex: No wad files found.");
		sprintf(wadfilename, "textures/%s", ff.cFileName);
		
		if (verbose)
			printf("Looking for texture %s...\n", miptex[i]);
		
		do 
		{
			missing_texture = false;

			strcpy(fullpath, wadfilename);
	
			if (verbose)
				printf("Searching %s...\n", fullpath);

			TEX_InitFromWad (fullpath);
			
			l->dataofs[i] = data - (byte *)l;
			len = LoadLump (miptex[i], data);
			if (data + len - dtexdata >= MAX_MAP_MIPTEX)
				Error ("Textures exceeded MAX_MAP_MIPTEX");
			if (!len || (strstr (miptex[i], "ignore"))) {// pOx - MAJOR HACK - ignore is no good for Software Quake
				l->dataofs[i] = -1;	// didn't find the texture
				missing_texture = true;
			}
			data += len;
			texdatasize = data - dtexdata;
			
			if (missing_texture)
			{
				if (!FindNextFile(ffhandle, &ff))
					ran_out = true;
				else
					sprintf(wadfilename, "textures/%s", ff.cFileName);
			}
	
		} while (missing_texture && !ran_out);
		
		if (missing_texture && ran_out) {
			
			if (strstr (miptex[i], "ignore"))
				printf ("WARNING: Omitted \"ignore\" texture lump (use with caution in Software Quake!)\n\n", miptex[i]);
			else
				printf ("WARNING: texture %s not found\n\n", miptex[i]);
			
		} else if (verbose)
			printf("Found texture %s, in wad %s\n\n", miptex[i], fullpath);
	}

}

//===========================================================================


/*
==================
BeginBSPFile
==================
*/
void BeginBSPFile (void)
{
// edge 0 is not used, because 0 can't be negated
	numedges = 1;

// leaf 0 is common solid with no faces
	numleafs = 1;
	dleafs[0].contents = CONTENTS_SOLID;

	firstface = 0;	
}


/*
==================
FinishBSPFile
==================
*/
void FinishBSPFile (void)
{
	printf ("--- FinishBSPFile ---\n");
	printf ("WriteBSPFile: %s\n\n", bspfilename);
	
	WriteMiptex ();

	PrintBSPFileSizes ();
	WriteBSPFile (bspfilename, bspversion);
}

